#!/usr/bin/env python
# programs a user image over serial

import cmd, getopt, os, subprocess, sys
try:
    student_mode = False
    from pmImgCreator import PmImgCreator
    import ipm
except ImportError:
    student_mode = True
    from rone_libs.pmImgCreator import PmImgCreator
    from rone_libs import ipm

__usage__ = """USAGE:
    robot.py -[d|s /dev/tty] --[desktop | serial=/dev/tty [baud=19200]]
             [ image.bin | main.py | [ fail | ipm | run ] ]

    -h          Prints this usage message.
    --help

    -s <port> [<baud>] Specifies the port (device) for a serial connection.
                <port> resembles `com5` on Win32 or `/dev/cu.usbmodem1912`.
                Optional argument, <baud>, defaults to 19200.

    --serial=<port> Specifies the port (device) for a serial connection.

    --baud=<baud>   Specifies the baud rate for a serial connection.

REQUIREMENTS:

    - pySerial package from http://pyserial.sourceforge.net/
    """

NEED_PYSERIAL = "Install the pySerial module from http://pyserial.sourceforge.net/"
REPLY_TERMINATOR  = '\x04'
BAUD = 230400

def U16_to_str(w):
    """Convert the 16-bit word, w, to a string of two bytes.

    The 2 byte string is in little endian order.
    DOES NOT INSERT TYPE BYTE.
    """

    return chr(w & 0xff) + \
           chr((w >> 8) & 0xff)


class AutodetectError(Exception): pass
def autodetect():
    if 'arwin' in sys.platform or 'linux' in sys.platform:
        devs = os.listdir('/dev/')

        if 'arwin' in sys.platform:
            devs = [d for d in devs if 'tty.usbserial' in d]
        else:
            devs = [d for d in devs if 'ttyUSB' in d]

        if len(devs) > 1:
            raise AutodetectError, "multiple FTDI devices found"
        if len(devs) < 1:
            raise AutodetectError, "no FTDI devices found"
        else:
            dev = '/dev/' + devs[0]
            return dev
    else:
        raise AutodetectError, "cannot autodetect on platform %s" % sys.platform


class Programmer(cmd.Cmd):
    def __init__(self, conn):
        cmd.Cmd.__init__(self,)
        self.conn = conn


    def send_command(self, cmd):
        self.conn.write(cmd)


    def program(self, fnames):
        if fnames[0][-2:] == 'py':
            print "compiling %s." % list(fnames)
            codeimg = compile_image(fnames)
        else:
            print "flashing raw image: %s" % fnames[0]
            codeimg = open(fnames[0]).read()

        # include name and null terminator in code image
        # this is how pmImgCreator does things, so we should match
        fn = os.path.basename(fnames[0])
        mod_name = os.path.splitext(fn)[0]

        if not (0 < len(mod_name) < 254):
            print "ERROR: length of primary module (%s) must be between 1-255 characters" % mod_name
            sys.exit()
        
        # include offset of start of codeimg,
        # null-terminated module name and null-terminated code image
        codeimg = chr(len(mod_name)+2) + mod_name + '\x00' + codeimg + '\x00'

        # convert to hex and add header/footer
        expected_sum = reduce(lambda x,y: x ^ ord(y), codeimg, 0)
        print "using %d bytes of flash on target" % len(codeimg)
        codeimg = 'l' + ipm.tohex(codeimg) + 'xx'

        print "length of encoded payload = ", len(codeimg)
        # print "CODEIMG: %s" % codeimg

        try:
            self.conn.write(codeimg)
            flash_response = self.conn.s.read(1)
            if not flash_response:
                print "ERROR: programming failed, response timed out"

            if flash_response == 'e':
                print "WARNING: programming succeeded after retries."
            elif flash_response == 'f':
                print "ERROR: programming failed after too many retries."
            elif flash_response == 's':
                print "flash succeeded."
            else:
                print "ERROR: programming failed (%s)" % flash_response

            reported_sum = int(self.conn.s.read(2), 16)
            if not reported_sum == expected_sum:
                print "ERROR: programming failed, incorrect checksum"
            else:
                print "programming verified!"
        except Exception, e:
            sys.stdout.write(
                "Connection write error, type Ctrl+%s to quit.\n" % EOF_KEY)


def compile_image(fnames):
    pic = PmImgCreator()
    pic.set_options('no_output_file', '.bin', 'usr', 'ram', None, fnames)
    pic.convert_files()
    
    codeobj = pic.formatFromExt[pic.imgtype]()
    return codeobj


def parse_cmdline():
    """Parses the command line for options.
    """
    baud = BAUD
    try:
        serdev = autodetect()
        Conn = ipm.SerialConnection
    except AutodetectError:
        serdev = None

    try:
        opts, args = getopt.getopt(sys.argv[1:], "dhs",
            ["desktop", "help", "serial=", "baud="])
    except Exception, e:
        print __usage__
        sys.exit()

    if not opts and not serdev:
        print __usage__
        sys.exit()

    for opt in opts:
        if opt[0] == "-s":
            Conn = ipm.SerialConnection
            serdev = args[0]
            if len(args) > 1:
                baud = int(args[1])
        elif opt[0] == "--serial":
            Conn = ipm.SerialConnection
            serdev = opt[1]
        elif opt[0] == "--baud":
            assert serdev, "--serial must be specified before --baud."
            baud = int(opt[1])
        else:
            print __usage__
            sys.exit(0)

    if Conn == ipm.SerialConnection:
        c = Conn(serdev, baud)
    else:
        c = Conn()

    return c, args

def do_args(conn, prog, args):
    if args[0] == 'ipm':
        prog.send_command('i')
        i = ipm.Interactive(conn)
        i.run()
    elif args[0] == 'listen':
        conn.read()
    elif args[0] == 'run':
        prog.send_command('r')
        conn.read()
    elif args[0] == 'echo':
        prog.send_command('e')
    elif args[0] == 'fail':
        prog.send_command('x')
    else:
        prog.program(args)

def init_ipm(serdev):
    conn = ipm.SerialConnection(serdev, BAUD)
    p = Programmer(conn)
    p.send_command('i')
    i = ipm.Interactive(conn)
    return i

def run(serdev, *args):
    conn = ipm.SerialConnection(serdev, BAUD)
    p = Programmer(conn)
    do_args(conn, p, list(args))

def connect(serdev):
    conn = ipm.SerialConnection(serdev, BAUD)
    p = Programmer(conn)
    do_args(conn, p, ['ipm'])    

def program(serdev, *args):
    conn = ipm.SerialConnection(serdev, BAUD)
    p = Programmer(conn)
    do_args(conn, p, list(args))

def main():
    conn, args = parse_cmdline()
    p = Programmer(conn)
    do_args(conn, p, args)

if __name__ == "__main__":
    main()
