#!/usr/bin/env python

# This file is Copyright 2007, 2009 Dean Hall.
#
# This file is part of the Python-on-a-Chip program.
# Python-on-a-Chip is free software: you can redistribute it and/or modify
# it under the terms of the GNU LESSER GENERAL PUBLIC LICENSE Version 2.1.
# 
# Python-on-a-Chip is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# A copy of the GNU LESSER GENERAL PUBLIC LICENSE Version 2.1
# is seen in the file COPYING in this directory.

"""
==================
Interactive PyMite
==================

An interactive command line that runs on a host computer that is connected
to a target device that is running PyMite.  The host computer compiles the
interactive statement and converts it to a form that PyMite can handle,
sends that over the connection where the target device loads and interprets it.
The target device then packages any result, sends it to the host computer
and the host computer prints the result.
"""

## @file
#  @copybrief ipm

## @package ipm
#  @brief Interactive PyMite
#
#
#  An interactive command line that runs on a host computer that is connected
#  to a target device that is running PyMite.  The host computer compiles the
#  interactive statement and converts it to a form that PyMite can handle,
#  sends that over the connection where the target device loads and interprets it.
#  The target device then packages any result, sends it to the host computer
#  and the host computer prints the result.

import cmd, dis, getopt, os, subprocess, sys, datetime
import pmImgCreator

try:
    import tool_info
    has_tool_info = True
except ImportError:
    has_tool_info = False

__fileids__ = {0x01:  'codeobj.c',
               0x02:  'dict.c',
               0x03:  'frame.c',
               0x04:  'func.c',
               0x05:  'global.c',
               0x06:  'heap.c',
               0x07:  'img.c',
               0x08:  'int.c',
               0x09:  'interp.c',
               0x0A:  'pmstdlib_nat.c',
               0x0B:  'list.c',
               0x0C:  'main.c',
               0x0D:  'mem.c',
               0x0E:  'module.c',
               0x0F:  'obj.c',
               0x10:  'seglist.c',
               0x11:  'sli.c',
               0x12:  'strobj.c',
               0x13:  'tuple.c',
               0x14:  'seq.c',
               0x15:  'pm.c',
               0x16:  'thread.c',
               0x17:  'float.c',
               0x18:  'class.c'}

__errors__ = {0xFF: 'No result',
              0xFE: 'General failure',
              0xFD: 'Return value for stub function',
              0xFC: 'Assertion failure',
              0xFB: 'Modified frame pointer',
              0xE0: 'General exception',
              0xE1: 'System exit',
              0xE2: 'Input/output error',
              0xE3: 'Division by zero',
              0xE4: 'Assertion error',
              0xE5: 'Attribute error',
              0xE6: 'Import error',
              0xE7: 'Index error',
              0xE8: 'Key error',
              0xE9: 'Memory error',
              0xEA: 'Name error',
              0xEB: 'Syntax error',
              0xEC: 'System error',
              0xED: 'Type error',
              0xEE: 'Value error',
              0xEF: 'Stop iteration',
              0xF0: 'Warning',
              0xF1: 'rone error',
              0xF2: 'Unbound local error'}

__eKeys__ = {0: 'Error      ',
             1: 'Release    ',
             2: 'File       ',
             3: 'Line       ',
             4: 'Python File',
             5: 'Python Line',
             6: 'Info',
             7: 'Traceback (top first)'}

__eVals__ = {0: __errors__,
             2: __fileids__}

__eFormatters__ = {
    0xEA: ": name '%s' is not defined",
    0xE5: ": module does not have attribute '%s'",
    0xE8: ": key '%s' not found"}


__usage__ = """USAGE:
    ipm.py -[d|s /dev/tty] --[desktop | serial=/dev/tty [baud=19200]]

    -h          Prints this usage message.
    --help

    -d          Specifies a desktop connection; uses pipes to send/receive bytes
    --desktop   to/from the target, which is the vm also running on the desktop.
                ipm spawns the vm and runs ipm-desktop as a subprocess.

    -s <port> [<baud>] Specifies the port (device) for a serial connection.
                <port> resembles `com5` on Win32 or `/dev/cu.usbmodem1912`.
                Optional argument, <baud>, defaults to 19200.

    --serial=<port> Specifies the port (device) for a serial connection.

    --baud=<baud>   Specifies the baud rate for a serial connection.

REQUIREMENTS:

    - pySerial package from http://pyserial.sourceforge.net/
    """

NEED_PYSERIAL = "Install the pySerial module from http://pyserial.sourceforge.net/"
if not sys.platform.lower().startswith("win"):
    PMVM_EXE = "../platform/desktop/main.out"
else:
    PMVM_EXE = "../platform/windows/Debug/pymite.exe"
IPM_PROMPT = "robot> "
COMPILE_FN = "<interactive>"
COMPILE_MODE = "single"
INIT_MESSAGE = """Welcome to Rice Python-on-a-Robot!
"""
HELP_MESSAGE = """
Type the Python code that you want to run on the robot.
If you see no prompt, type two consecutive returns to exit multiline mode.
Type Ctrl+C to interrupt and Ctrl+D to quit.
"""

REPLY_TERMINATOR  = '\x04'
GROUP_SEPARATOR   = '\x1d'
RECORD_SEPARATOR  = '\x1e'
UNIT_SEPARATOR    = '\x1f'

if sys.platform.lower().startswith("win"):
    EOF_KEY = 'Z'
else:
    EOF_KEY = 'D'


def tohex(codeimg):
    out = []
    for byte in codeimg:
        unformatted = hex(ord(byte))
        formatted = unformatted.split('x')[1]
        formatted = '0'*(2-len(formatted)) + formatted

        out.append(formatted)

    return ''.join(out)


class PyMiteError(Exception):
    pass


class Connection(object):
    def open(self,): raise NotImplementedError
    def read(self,): raise NotImplementedError
    def write(self, msg): raise NotImplementedError
    def close(self,): raise NotImplementedError


class PipeConnection(Connection):
    """Provides ipm-host to target connection over stdio pipes on the desktop.
    This connection should work on any POSIX-compliant OS.
    The ipm-device must be spawned as a subprocess
    (the executable created when PyMite was built with PLATFORM=desktop).
    """
    def __init__(self, target=PMVM_EXE):
        self.open(target)


    def open(self, target):
        self.child = subprocess.Popen(target,
                                      bufsize=-1,
                                      stdin=subprocess.PIPE,
                                      stdout=subprocess.PIPE,
                                      stderr=subprocess.PIPE,
                                      )


    def read(self,echo=True):
        response = ""
        in_exception = False

        while True:
            character = self.child.stdout.read(1)
            if character == '':
                break

            response += character

            if character == REPLY_TERMINATOR:
                return response

            if character == GROUP_SEPARATOR:
                in_exception = (not in_exception)
                continue

            if not in_exception:
                sys.stdout.write(character)


    def write(self, msg):
        self.child.stdin.write(msg)
        self.child.stdin.flush()


    def close(self,):
        self.write("\0")


class SerialConnection(Connection):
    """Provides ipm-host to target connection over a serial device.
    This connection should work on any platform that PySerial supports.
    The ipm-device must be running at the same baud rate (19200 default).
    """

    def __init__(self, serdev="/dev/cu.SLAB_USBtoUART", baud=19200):
        try:
            import serial
        except Exception, e:
            print NEED_PYSERIAL
            raise e


        self.s = serial.Serial(serdev, baud)
        self.s.setRtsCts(1)
        self.s.setTimeout(4)

    def read(self, echo=True):
        response = ""
        in_exception = False

        while True:
            character = self.s.read(1)
            response += character

            if character == REPLY_TERMINATOR:
                return response

            if character == GROUP_SEPARATOR:
                in_exception = (not in_exception)
                continue

            if (not in_exception) and echo:
                sys.stdout.write(character)
                

#    def read(self,):
#        # Collect all characters up to and including the ipm reply terminator
#        line = ""
#        while ((len(line) == 0) or (line[-1] != REPLY_TERMINATOR)):
#            line += self.s.readline(eol=REPLY_TERMINATOR)
#        return line


    def write(self, msg):
        self.s.write(msg)
        self.s.flush()
            
    
    def close(self,):
        self.s.close()


class Interactive(cmd.Cmd):
    """The interactive command line parser accepts typed input line-by-line.
    If a statement requires multiple lines to complete,  the input
    is collected until two sequential end-of-line characters are received.
    """
    ipmcommands = ("?", "help", "run", "verbose")


    def __init__(self, conn):
        cmd.Cmd.__init__(self,)
        self.prompt = IPM_PROMPT
        self.conn = conn
        conn.stdout = self.stdout
        self.stderr = sys.stderr
        self.pic = pmImgCreator.PmImgCreator()
        self.last_err = None

        # open the log file
        self.logfile = open('ipm-errors.log', 'a')


    def do_help(self, *args):
        """Prints the help message.
        """
        self.stdout.write(HELP_MESSAGE)

    
    def do_verbose(self, *args):
        if self.last_err:
            self.report_error(self.last_err, verbose=True)


    def do_run(self, *args):
        """runs a module from the host to the target device.
        """

        # Ensure the filename arg names a python source file
        fn = args[0]
        if not fn.endswith(".py"):
            self.stderr.write('Error using "run <module>": '
                              'module must be a ".py" source file.\n')
            return

        try:
            file = open(fn)
        except IOError:
            self.stderr.write("Error using run: file not found")

        try:
            codeobj = compile(file.read(), fn, 'exec')
        except Exception, e:
            self.stderr.write("%s:%s\n" % (e.__class__.__name__, e))
            return
        
        self.run_codeobj(codeobj)

    def eval_string(self, cmdstr):
        """Executes cmdstr
        """
        # Try to compile the given line
        codeobj = None
        try:
            codeobj = compile(cmdstr, COMPILE_FN, COMPILE_MODE)
        except SyntaxError, se:
            self.stderr.write("%s:%s\n" % (se.__class__.__name__, se))
            return

        # Run the compiled code object
        return self.run_codeobj(codeobj, fail_on_error=True, echo=False)

    def onecmd(self, line):
        """Gathers one interactive line of input (gets more lines as needed).
        """
        # Ignore empty line, continue interactive prompt
        if not line:
            return

        # Handle ctrl+D (End Of File) input, stop interactive prompt
        if line == "EOF":
            self.conn.close()

            # Do this so OS prompt is on a new line
            self.stdout.write("\n")

            # Quit the run loop
            self.stop = True
            return True

        # Handle ipm-specific commands
        if line.split()[0] in Interactive.ipmcommands:
            cmd.Cmd.onecmd(self, line)
            return

        # Gather input from the interactive line
        codeobj = None
        while not codeobj:

            # Try to compile the given line
            try:
                codeobj = compile(line, COMPILE_FN, COMPILE_MODE)

            # Get more input if syntax error reports unexpected end of file
            except SyntaxError, se:

                # Print any other syntax error
                if not se.msg.startswith("unexpected EOF while parsing"):
                    self.stderr.write("%s:%s\n" % (se.__class__.__name__, se))
                    return

                # Restore the newline chopped by cmd.py:140
                line += "\n"

                # Get more input if needed
                while not line.endswith("\n\n"):
                    line += self.stdin.readline()

            # Print any other exception
            except Exception, e:
                self.stderr.write("%s:%s\n" % (e.__class__.__name__, e))
                return
            
        self.run_codeobj(codeobj)


    def run_codeobj(self, codeobj, fail_on_error=False, echo=True):
        # DEBUG: Uncomment the next line to print the statement's bytecodes
        # dis.disco(codeobj)

        # Convert to a code image
        try:
            codeimg = self.pic.co_to_str(codeobj)
            
        # Print any conversion errors
        except Exception, e:
            self.stderr.write("%s:%s\n" % (e.__class__.__name__, e))

        # Otherwise send the image and print the reply
        else:

            # DEBUG: Uncomment the next line to print the size of the code image
            # print "DEBUG: len(codeimg) = ", len(codeimg)
            # DEBUG: Uncomment the next line to print the code image
            # print "DEBUG: codeimg = ", repr(codeimg)
            try:
                hexstr = 'c' + tohex(codeimg)
                self.conn.write(hexstr)
            except Exception, e:
                self.stderr.write(
                    "Connection write error, type Ctrl+%s to quit.\n" % EOF_KEY)

            rv = self.conn.read(echo=echo)
            if rv == '':
                self.stderr.write(
                    "Connection read error, type Ctrl+%s to quit.\n" % EOF_KEY)
            #elif rv[0] == UNIT_SEPARATOR:
            #    self.report_error(rv)
            else:
                # print "DEBUG: output from target: %s" % repr(rv)
                clean_output = rv.strip('\x04').strip('\n') + '\n'

                if not clean_output.strip():
                    clean_output = ''

                err_output = clean_output.split(GROUP_SEPARATOR)
                # print err_output

                # Normal text output (before any exception report)
                # self.stdout.write(err_output[0])

                output = err_output[0]

                if len(err_output) > 1:
                    # Exception report
                    try:
                        self.report_error(err_output[1])
                    except KeyError:
                        print "error decoding new style exception, printing what's available"
                        pairs = err_output[1].split(RECORD_SEPARATOR)
                        for pair in pairs:
                            print pair.split(UNIT_SEPARATOR)

                    if fail_on_error:
                        raise PyMiteError

                if len(err_output) > 2:
                    output += err_output[2]
                    # Normal text after exception report
                    # self.stdout.write(err_output[2])

                return output


    def run(self,):
        """Runs the command loop and handles keyboard interrupts (ctrl+C).
        The command loop is what calls self.onecmd().
        """

        print INIT_MESSAGE,

        if has_tool_info:
            print "Running tools version: %s" % tool_info.revision
        else:
            print "WARNING: Running working copy of tools."
        try:
            self.eval_string('import build')
            version = self.eval_string('build.revision').strip()
            print "Connected to robot. Software version: %s" % version
            if 'M' in version or ':' in version:
                print "WARNING! Running development robot software. Contact TA."
        except PyMiteError:
            print "WARNING! No build info module. Running ancient software or IPM?"

        print HELP_MESSAGE,

        self.stop = False
        while not self.stop:
            try:
                self.cmdloop()
            except KeyboardInterrupt, ki:
                print "\n", ki.__class__.__name__
                # TODO: check connection?


    def report_error(self, err, verbose=False):
        self.last_err = err
        pairs = err.split(RECORD_SEPARATOR)
        unpacked_pairs = [p.split(UNIT_SEPARATOR) for p in pairs]
        decoded_pairs = [(int(key, 16), val) for (key, val) in unpacked_pairs]

        exception_report = dict(decoded_pairs)


        if verbose:
            for (key, val) in decoded_pairs:
                if key in __eVals__:
                    print >> sys.stderr, __eKeys__[key], ":", __eVals__[key][int(val, 16)]
                else:
                    print >> sys.stderr, __eKeys__[key], ":", val
        else:
            error_code = int(exception_report[0], 16)
            name = __errors__[error_code]
            c_file = exception_report[2]
            c_line = exception_report[3]
            python_file = exception_report[4]
            python_line = exception_report[5]
            info = exception_report[6]

            if error_code in __eFormatters__.keys():
                print >> sys.stderr, name + (__eFormatters__[error_code] % info)
            elif info:
                print >> sys.stderr, "%s: %s" % (name, info)
            else:
                print >> sys.stderr, name



            print >> sys.stderr, "File %s, line %s" % (python_file, python_line)

            # log the exception
            time = datetime.datetime.now().isoformat()
            log_entry =  "/".join((name, c_file, c_line, python_file, \
                                   python_line, info, time))

            self.logfile.write(log_entry + '\n')
            self.logfile.flush()

def parse_cmdline():
    """Parses the command line for options.
    """
    baud = 19200
    Conn = PipeConnection
    serdev = None

    try:
        opts, args = getopt.getopt(sys.argv[1:], "dhs",
            ["desktop", "help", "serial=", "baud="])
    except Exception, e:
        print __usage__
        sys.exit()

    if not opts:
        print __usage__
        sys.exit()

    for opt in opts:
        if opt[0] == "-d" or opt[0] == "--desktop":
            Conn = PipeConnection
        elif opt[0] == "-s":
            Conn = SerialConnection
            serdev = args[0]
            if len(args) > 1:
                baud = int(args[1])
        elif opt[0] == "--serial":
            Conn = SerialConnection
            serdev = opt[1]
        elif opt[0] == "--baud":
            assert serdev, "--serial must be specified before --baud."
            baud = int(opt[1])
        else:
            print __usage__
            sys.exit(0)

    if Conn == SerialConnection:
        c = Conn(serdev, baud)
    else:
        c = Conn()

    return c


def main():
    conn = parse_cmdline()
    i = Interactive(conn)
    i.run()


def ser_test():
    """Test ipm over serial connection directly.
    """
    try:
        import serial
    except Exception, e:
        print NEED_PYSERIAL
        raise e

    pic = pmImgCreator.PmImgCreator()
    serconn = serial.Serial("/dev/cu.SLAB_USBtoUART", 19200)
    serconn.setTimeout(2)

    testcode = (
        'print "Hello"\n',
        'import sys\n',
        'print sys.heap()\n',
        )

    for line in testcode:
        print "compiling ``%s``" % line
        codeobj = compile(line, COMPILE_FN, COMPILE_MODE)
        codeimg = pic.co_to_str(codeobj)
        print "codeimg is %d bytes" % len(codeimg)
        print "sending codeimg..."
        serconn.write(codeimg)
        reply = serconn.readline(eol=REPLY_TERMINATOR)
        print "reply is %d bytes" % len(reply)
        print "reply is:\n%s" % reply


if __name__ == "__main__":
    main()
